package org.torproject.metrics.descriptorparser.parsers;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.fail;

import org.torproject.metrics.descriptorparser.utils.PsqlConnector;

import org.junit.Before;
import org.junit.Test;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.Timestamp;

public class RouterStatusBuilderTest {
  /**
   * Insert server descriptor data into the database before doing tests.
   */
  @Before()
  public void insertServerDescriptorsAndBuildStatuses() throws Exception {
    String serverPath =
        "src/test/resources/2022-08-31-12-05-00-server-descriptors";
    String confFile = "src/test/resources/config.properties.test";
    String serverDigest = "e93044e24f71bda8595907f0429f009573e5a140";
    String fingerprint = "22F74E176F803499D4F80D9CE7D325883A8C0E45";

    Connection conn = null;
    PsqlConnector psqlConn = new PsqlConnector();
    conn = psqlConn.connect(confFile);

    ServerDescriptorParser ep = new ServerDescriptorParser();
    ep.run(serverPath, conn);
  }

  @Test()
  public void testRouterStatusBuilder() throws Exception {
    String confFile = "src/test/resources/config.properties.test";
    String fingerprint = "8D896C8B367813030591A00DB7E7722EF6C4C23C";

    Connection conn = null;
    PsqlConnector psqlConn = new PsqlConnector();
    conn = psqlConn.connect(confFile);

    PreparedStatement preparedStatement = conn.prepareStatement(
        "SELECT * FROM server_status WHERE fingerprint = '"
        + fingerprint + "' LIMIT 1;");

    try (ResultSet rs = preparedStatement.executeQuery()) {
      if (rs.next()) {
        assertEquals(rs.getString("nickname"), "Luxembourg");
        assertEquals(rs.getString("fingerprint"),
            fingerprint);
        assertEquals(rs.getString("or_addresses"),
            "[[2605:6400:30:f82b:c986:c57f:3adc:b9da]:443]");
        assertEquals(rs.getString("unreachable_or_addresses"), "");
      } else {
        fail("Status not found");
      }
    }

  }

  @Test()
  public void testInsertingStatusesAtDifferentTime() throws Exception {
    Connection conn = null;
    PsqlConnector psqlConn = new PsqlConnector();
    String confFile = "src/test/resources/config.properties.test";
    conn = psqlConn.connect(confFile);

    String consensusPathSt =
        "src/test/resources/2023-02-05-20-00-00-consensus";
    String consensusPathNd =
        "src/test/resources/2023-02-02-00-00-00-consensus";
    ConsensusParser cp = new ConsensusParser();
    cp.run(consensusPathSt, conn);
    cp.run(consensusPathNd, conn);

    String serverPath = "src/test/resources/server-descriptors-through-time";
    ServerDescriptorParser sp = new ServerDescriptorParser();
    sp.run(serverPath, conn);

    String fingerprint = "22F74E176F803499D4F80D9CE7D325883A8C0E45";
    PreparedStatement preparedStatement = conn.prepareStatement(
        "SELECT * FROM server_status WHERE fingerprint = '"
        + fingerprint + "' AND published<='2023-02-05 12:00:02' ORDER BY"
        + " published DESC LIMIT 1;");

    try (ResultSet rs = preparedStatement.executeQuery()) {
      if (rs.next()) {
        assertEquals(rs.getString("nickname"), "MakeSecure");
        assertEquals(rs.getString("fingerprint"),
            fingerprint);
        assertEquals(rs.getTimestamp("last_seen"),
            Timestamp.valueOf("2023-02-05 12:00:02"));
      } else {
        fail("Status not found");
      }
    }

  }

  @Test()
  public void testInsertingBridgesStatuses() throws Exception {
    String confFile = "src/test/resources/config.properties.test";
    Connection conn = null;
    PsqlConnector psqlConn = new PsqlConnector();
    conn = psqlConn.connect(confFile);

    String consensusPathForVersions =
        "src/test/resources/2022-08-30-10-00-00-consensus";
    ConsensusParser cp = new ConsensusParser();
    cp.run(consensusPathForVersions, conn);

    String serverPath =
        "src/test/resources/bridges-server-desc";
    ServerDescriptorParser sp = new ServerDescriptorParser();
    sp.run(serverPath, conn);

    String fingerprint = "00B5DE0522A6F215BF694DE36576AF1A3927D6D8";
    PreparedStatement preparedStatement = conn.prepareStatement(
        "SELECT * FROM server_status WHERE fingerprint='"
        + fingerprint + "' AND published<='2022-08-30 23:54:51' ORDER BY"
        + " published DESC LIMIT 1;");

    try (ResultSet rs = preparedStatement.executeQuery()) {
      if (rs.next()) {
        assertEquals(rs.getString("nickname"), "wolf359");
        assertEquals(rs.getString("fingerprint"),
            fingerprint);
        assertEquals(rs.getTimestamp("last_seen"),
            Timestamp.valueOf("2022-08-30 11:54:43"));
      } else {
        fail("Status not found");
      }
    }
  }

  @Test()
  public void testRouterFamily() throws Exception {
    String confFile = "src/test/resources/config.properties.test";
    Connection conn = null;
    PsqlConnector psqlConn = new PsqlConnector();
    conn = psqlConn.connect(confFile);
    String fingerprint = "E42693F28ECCB5AFBC290371DF275C6FD0657F78";

    PreparedStatement preparedStatement = conn.prepareStatement(
        "SELECT * FROM server_status WHERE fingerprint = '"
        + fingerprint + "' LIMIT 1;");

    try (ResultSet rs = preparedStatement.executeQuery()) {
      if (rs.next()) {
        assertEquals(rs.getString("fingerprint"), fingerprint);
        assertEquals(rs.getString("nickname"), "amatista233");
        String family = "[\"$043BC8636292695FF268455FDDE068D510A4E6EA\","
            + "\"$18EC7D1CBB8A22AED6A7408FDEDF985AC099C9D4\","
            + "\"$1F9BEE841B906441D255FE43EA4E1156F0E61188\","
            + "\"$42E360F5FCA139ADD7C01BC706D29FE617026F74\","
            + "\"$46D5AA822C1487262387B87348C3EAE6F940BEC9\","
            + "\"$580A37D555E016CD14AB8C471E6E446FEBC3D804\","
            + "\"$6AF21ACE616EC712110388B2134CC1A955180227\","
            + "\"$6DA9ECDE07B779B325A5A7275F8038E8F72A083F\","
            + "\"$6EECA1B9DD50FC69FC6432A1F7A2AD26C1212893\","
            + "\"$7282D0E6FB01E2B4BE550CB4EA433AE16DB65C17\","
            + "\"$794E6B42020F6F26286DC41B33A90BA1ED0DA710\","
            + "\"$7BD0DE299AFDF4DD8E1A82472FBC6F8015228B0B\","
            + "\"$7EECF9ED6303D79871C05669646922E16B23C090\","
            + "\"$95F9DD417C8A4FA717FC7D370389C345BF12A2EA\","
            + "\"$A227942C424EC58CBCB63580541EF764A8460B03\","
            + "\"$A44CA851FF0B01D8B76856C3A5140426E7CC5024\","
            + "\"$A4E0F37BAC3E38F503B42BC19B3C0BFADF8248DD\","
            + "\"$D2A06E75132502D4C318BA9F6A44099F485FD4EC\","
            + "\"$D59B3934082BAE3A653697094E80AA295E733952\","
            + "\"$DBBA116A7B97E58D8837D467F818270B1503CADC\","
            + "\"$E42693F28ECCB5AFBC290371DF275C6FD0657F78\","
            + "\"$E5F513E15CC2CD35DFC4ECEAD190D3B38DB61B9E\","
            + "\"$E92EAB61F50B40D935EB943910700E4C44DCA36E\","
            + "\"$EAF570DAEFA4CFEE04FE9AE3747CC9AC70B23A59\","
            + "\"$EEB1C2EC72096E79F5583B74074A16D14D5CD10B\","
            + "\"$FBC2D4967A73C8962ED7769F4F421C5D3EC9B594\"]";
        assertEquals(rs.getString("declared_family"), family);
        String effFamily = "[\"A4E0F37BAC3E38F503B42BC19B3C0BFADF8248DD\","
            + "\"E42693F28ECCB5AFBC290371DF275C6FD0657F78\"]";
        assertEquals(rs.getString("effective_family"), effFamily);
      } else {
        fail("Status not found");
      }
    }
  }

}
